
import os
from flask import Flask, flash, request, redirect, url_for, render_template, session, g
from werkzeug.utils import secure_filename

import numpy as np
import plotly.io as pio
import plotly.graph_objects as go

import graficas as gph
import amodal

app = Flask(__name__)
global upload_folder
upload_folder = '/home/jesus/1tfg/codigo/Programa_Pycharm/FlaskFolder/static'
app.config['UPLOAD_FOLDER'] = os.path.join(upload_folder,'upload_xlsx')
app.config['SIGNAL_FOLDER'] = os.path.join(upload_folder,'db_tiempo')
app.config['FRF_FOLDER'] = os.path.join(upload_folder,'db_frecuencia')

app.secret_key = 'analisismodal'

user_key = 0
def key_generator():
    global user_key
    if user_key < 100:
        user_key +=1
    else:
        user_key = 0
    session['userkey'] = str(user_key)
    session['signalpath'] = os.path.join(app.config['SIGNAL_FOLDER'],str(user_key))
    session['frfpath'] = os.path.join(app.config['FRF_FOLDER'],str(user_key))



@app.route('/')
def info():
    return render_template('info.html')


@app.route('/cargar_datos', methods=['GET', 'POST'])
def upload_file():

    filename = 'No hay archivos'
    presetname = 'No hay archivos'
    if request.method == 'POST':
        # check if the post request has the file part
        if 'file' in request.files:

            file = request.files['file']
            if file.filename == '':
                flash('No se han seleccionado datos')
                return redirect(request.url)

            filename = secure_filename(file.filename)
            print(filename)

            file.save(os.path.join(app.config['UPLOAD_FOLDER'], filename))
            session['filename']=filename
            session['fileroute'] = os.path.join(app.config['UPLOAD_FOLDER'], filename)
            key_generator()

            senal = amodal.Tiempo()
            senal.lectura(session['fileroute'])
            senal.calcular()
            senal.get_h5py(session['signalpath'])

            return redirect(request.url)

    try:
        filename = session['filename']
    except:
        pass
    try:
        presetname = session['presetname']
    except:
        pass


    return render_template('/cargar_datos.html',datos= filename)

@app.route('/senal', methods=['GET','POST'])
def senal():
    leyenda= np.zeros(2)
    rms=np.zeros(2)
    mtvv=np.zeros(2)
    vdv=np.zeros(2)
    fig = go.Figure()

    if request.method == 'POST':
        try:
            _ = session['userkey']
        except:
            flash('Cargue unos datos')
            return redirect(request.url)

        if 'aceleracion' not in request.form and 'fuerza' not in request.form and 'rms' not in request.form and 'vdv' not in request.form:
            flash('Datos incorrectos')
            return redirect(request.url)

        senal = amodal.Tiempo()
        senal.from_h5py(session['signalpath'])
        senal.calcular()
        arguments = list()
        if 'aceleracion' in request.form:
            print('aceleracion')
            arguments.append(senal.aceleracion)

        if 'fuerza' in request.form:
            print('fuerza')
            arguments.append(senal.fuerza)

        if 'rms' in request.form:
            print('rms')
            senal.get_rms_movil(avance_rms= 1)
            arguments.append(senal.rms_movil)

        if 'vdv' in request.form:
            print('vdv')
            arguments.append(senal.vdv)
        print(arguments[0].v)
        fig = gph.grafica(senal.tiempo,*arguments)

        leyenda = senal.aceleracion.l
        rms = np.round(senal.rms.v,5)
        mtvv = np.round(senal.mtvv.v,5)
        vdv = np.round(senal.vdv.v[:,-1],5)

    fig = pio.to_html(fig)
    return render_template('senal.html',grafica= fig, leyenda= leyenda, rms= rms,mtvv= mtvv, vdv= vdv )

@app.route('/tratamiento', methods=['GET', 'POST'])
def tratamiento():
    if request.method == 'POST':
        try:
            senal = amodal.Tiempo()
            senal.from_h5py(session['signalpath'])
        except:
            flash('Cargue unos datos')
            return redirect(request.url)

        detrend, resample, filtro = False, False, False
        if 'detrend' in request.form:
            detrend = True

        if 'samplerate' in request.form:
            try:
                amodal.preset.samplerate_resample = int(request.form['samplerate'])
            except:
                flash('Datos incorrectos')
                return redirect(request.url)
            resample = True

        if 'orden' in request.form:
            try:
                orden = int(request.form['orden'])
                amodal.preset.orden_filtro = orden
                frecuencia = abs(float(request.form['frecuencia']))
                amodal.preset.frecuencia_filtro = frecuencia
            except:
                flash('Datos incorrectos')
                return redirect(request.url)
            filtro = True

        if detrend is True:
            senal = amodal.Tiempo()
            senal.from_h5py(session['signalpath'])
            senal.detrend()
            senal.get_h5py(session['signalpath'])

        if resample is True:
            senal = amodal.Tiempo()
            senal.from_h5py(session['signalpath'])
            senal.resample()
            senal.get_h5py(session['signalpath'])

        if filtro is True:
            senal = amodal.Tiempo()
            senal.from_h5py(session['signalpath'])
            print('El samplerate es ', senal.samplerate)
            senal.filtro()
            senal.get_h5py(session['signalpath'])

        if 'nfft' in request.form:

            if request.form['nfft'] == '' or request.form['nperseg'] == '' or request.form['noverlap'] == '':
                print('error1')
                return redirect(request.url)
            noverlap = int(request.form['noverlap'])
            nperseg = int(request.form['nperseg'])
            nfft = int(request.form['nfft'])
            if nfft < nperseg:
                print('error2')
                flash('Nperseg debe ser menor o igual que nfft')
                return redirect(request.url)
            if nperseg <= noverlap:
                print('error3')
                flash('Noverlap debe ser menor que nperseg')
                return redirect(request.url)

            print('nfft success')
            amodal.preset.nfft_frf = nfft
            amodal.preset.noverlap_frf = noverlap
            amodal.preset.nperseg_frf = nperseg
            senal = amodal.Tiempo()
            senal.from_h5py(session['signalpath'])
            frf = senal.get_frf()
            frf.integrar()
            frf.get_h5py(session['frfpath'])
            print(frf.magnitud.v[0,3])



    return render_template('tratamiento.html')

@app.route('/frf')
def frf():
    mag = go.Figure()
    nyq = go.Figure()

    try:
        frf = amodal.Frecuencia()
        frf.from_h5py(session['frfpath'])
        frfs = True
    except:
        print("-------FALLO--------")
        pass

    if frfs is True:
        mag = gph.bode(frf)
        nyq = gph.nyquist(frf)

    bodemagnitud = pio.to_html(mag)
    bodenyquist = pio.to_html(nyq)
    return render_template('frf.html',bodemagnitud= bodemagnitud ,bodenyquist= bodenyquist)

@app.route('/modal', methods=['GET','POST'])
def modal():
    leyenda = np.zeros(2)
    resonancia = np.zeros([2,2])
    amortiguamiento = np.zeros([2,2])
    resonanciamedia = np.zeros([2,2])
    amortiguamientomedio = np.zeros([2,2])
    num_modos = 1
    try:
        num_modos = session['num_modos']
    except:
        pass

    if request.method == 'POST':
        print(request.form)
        if 'num_modos' in request.form:
            try:
                session['num_modos'] =  int(request.form['num_modos'])
                return redirect(request.url)
            except:
                flash('Introduzca un numero entero')
                return redirect(request.url)

        if '1-0' in request.form:
            tamano = len(request.form)
            tamano = tamano-1

            if tamano%2 != 0:
                flash('Introduzca todos los elementos')
                return redirect(request.url)
            tamano = (tamano)/2
            freqs = np.zeros([int(tamano),2])
            lista = list(request.form)
            i = 0
            for item in lista[0:-1]:
                if '1-' in item:
                    try:
                        freq = abs(float(request.form[item]))
                        freqs[i, 0] = freq
                    except:
                        flash('Introduzca un numero valido')
                        return redirect(request.url)
                if '2-' in item:
                    try:
                        freq = abs(float(request.form[item]))
                        freqs[i, 1] = freq
                    except:
                        flash('Introduzca un numero valido')
                        return redirect(request.url)
                    i +=1

            print(freqs)
            frf = amodal.Frecuencia()
            frf.from_h5py(session['frfpath'])

            lista = list()
            for i in range(frf.axis_0):
                lista.append(freqs)
            frecuencia = np.array([*lista])
            print(frecuencia.shape)
            modos = amodal.Modal(frf,frecuencia)
            modos.frec_to_place()
            modos.circle_fit()
            modos.promedio()
            print('----------------')
            print(modos.frec_resonancia)
            print(modos.amortiguamiento)
            print(modos.frec_resonancia_media)
            print(modos.amortiguamiento_medio)
            print('----------------')
            amortiguamientomedio = np.round(modos.amortiguamiento_medio,5)
            resonanciamedia = np.round(modos.frec_resonancia_media,5)
            amortiguamiento = np.round(np.transpose(modos.amortiguamiento),5)
            resonancia = np.round(np.transpose(modos.frec_resonancia),5)
            leyenda = frf.real.l

    return render_template('modal.html', num_intervalos= range(num_modos),resonancia= resonancia,
                           amortiguamiento= amortiguamiento, resonanciamedia= resonanciamedia,
                           amortiguamientomedio= amortiguamientomedio, leyenda = leyenda)


if __name__ == '__main__':
    app.run(debug=True)


