import pandas as pd
import numpy as np

from scipy import signal, optimize
import math
import cmath
from scipy import interpolate

from os import path
import importlib
import h5py as h5
import os
import plotly.graph_objects as go

class Default:
    archivo = None

    @classmethod
    def get(cls, _archivo='preset.py'):
        if path.exists(_archivo):
            if _archivo[-3:] == '.py':
                _archivo = _archivo[:-3]
                cls.archivo = _archivo
                return None
        return ImportError

    @classmethod
    def set(cls):
        if cls.archivo is not None:
            global preset
            preset = importlib.import_module(cls.archivo)
            return None
        return ImportError

    @classmethod
    def restore(cls, _archivo = 'preset.py'):

        with open(_archivo, 'w') as p:
            p.write('''# 
#------------------------------------------------
# Plantilla para guardar los valores por defecto
#------------------------------------------------
# 
# Esta plantilla se debe editar siguiendo las sintaxis de 
# Python, cualquier grafia que no sea propia de Python 
# resultara en un error de programa. De la misma forma,
# si se cambia el nombre de cualquiera de las variables
# resultara en un error de la funcion que dependa de dicha
# variable. 
#

ventana_rms = {}
solape_rms = {}

incremento_vdv = {}

orden_filtro = {}
frecuencia_filtro = {}

samplerate_resample = {}

ventana_frf = '{}'
nperseg_frf = {}
noverlap_frf = {}
nfft_frf = {}    

threshold_frf = {}

'''.format(
                1, 0.5,
                1,
                4, 200,
                512,
                'boxcar', 16384, 0, 16384,
                50
            ))

    @classmethod
    def printsettings(cls, _archivo = 'plantilla_preset.py'):
        with open(_archivo,'w') as p:
            p.write('''# 
#------------------------------------------------
# Plantilla para guardar los valores por defecto
#------------------------------------------------
# 
# Esta plantilla se debe editar siguiendo las sintaxis de 
# Python, cualquier grafia que no sea propia de Python 
# resultara en un error de programa. De la misma forma,
# si se cambia el nombre de cualquiera de las variables
# resultara en un error de la funcion que dependa de dicha
# variable. 
#

ventana_rms = {}
solape_rms = {}

incremento_vdv = {}

orden_filtro = {}
frecuencia_filtro = {}

samplerate_resample = {}

ventana_frf = '{}'
nperseg_frf = {}
noverlap_frf = {}  
nfft_frf = {}

threshold_frf = {}    
            
'''.format(
                preset.ventana_rms,
                preset.solape_rms,

                preset.incremento_vdv,

                preset.orden_filtro,
                preset.frecuencia_filtro,

                preset.samplerate_resample,

                preset.ventana_frf,
                preset.nperseg_frf,
                preset.noverlap_frf,
                preset.nfft_frf,
                preset.threshold_frf
            ))

if Default.get() == ImportError:
    Default.restore()
    Default.get()
Default.set()

class Atributo:
    # Guarda la informacion en tres datos separados.
    def __init__(self, valor, leyenda, unidad):
        self.v = valor
        self.l = leyenda
        self.u = unidad

class Tiempo:
    def __init__(self, tiempo=None, unidad_tiempo=None, leyenda_tiempo=None, aceleracion=None, unidad_aceleracion=None,
                 leyenda_aceleracion=None, fuerza=None, unidad_fuerza=None, leyenda_fuerza=None,
                 rms=None, rms_movil=None,  tiempo_rms_movil=None, mtvv=None, vdv=None, tiempo_vdv=None, samplerate=None):
        # Se inician los datos de la clase:
        self.tiempo = Atributo(tiempo, leyenda_tiempo, unidad_tiempo)
        self.aceleracion = Atributo(aceleracion, leyenda_aceleracion, unidad_aceleracion)
        self.fuerza = Atributo(fuerza, leyenda_fuerza, unidad_fuerza)

        # Se halla el samplerate y se comprueba que los vectores son correctos

        try:
            self.comprobar_samplerate()
        except:
            pass

        try:
            self.axis_0, self.axis_1 = self.aceleracion.v.shape
        except:
            pass

        self.rms = Atributo(valor=rms, leyenda= leyenda_aceleracion, unidad=unidad_aceleracion)
        self.rms_movil = Atributo(valor=rms_movil, leyenda= leyenda_aceleracion, unidad=unidad_aceleracion)
        self.tiempo_rms_movil = Atributo(valor=tiempo_rms_movil, leyenda=leyenda_tiempo, unidad= unidad_tiempo )
        self.mtvv = Atributo(valor=mtvv, leyenda= leyenda_aceleracion, unidad=unidad_aceleracion)
        self.vdv = Atributo(valor=vdv, leyenda= leyenda_aceleracion, unidad=unidad_aceleracion)
        self.tiempo_vdv = Atributo(valor=tiempo_vdv, leyenda=leyenda_tiempo, unidad= unidad_tiempo )



    def comprobar_samplerate(self):
        # Se comprueba primero el samplerate de manera precisa, contando el numero
        # de iteraciones sobre el vector tiempo hasta alcanzar el valor 1.
        # En caso de ser un vector de longitud suficiente retorna el samplerate exacto, samples
        samples = 0
        t = 0
        while t < 1:
            samples += 1
            if samples == len(self.tiempo.v):
                break
            t = self.tiempo.v[samples]
            if t == 1:
                self.samplerate = samples
        # Si el metodo anterior falla se recurre a hallar el samplerate dividiendo 1 entre el incremento
        # de tiempo dt = tiempo[1] - tiempo [0]. Si falla se retorna el valor más cercano al samplerate
        # con una indicación de error para que el usuario revise los datos

        samplerate_decimal = 1 / (self.tiempo.v[1] - self.tiempo.v[0])
        samplerate_round = round(samplerate_decimal, 6)
        samplerate = int(samplerate_round)

        self.samplerate = samplerate

    def comprobar_datos(self):

        # Se comprueba si la longitud de los vectores es coherente
        if len(self.tiempo.v) != len(self.fuerza.v) or len(self.aceleracion.v[0]) != len(self.tiempo.v):
            print('Error: la longitud de los vectores no coincide, revise los datos')

    def lectura(self, archivo, numero_hoja=1):
        df = pd.read_excel(archivo, sheet_name=numero_hoja)  # importa los datos como dataframe
        unidades = df[0:1].to_numpy()  # copia la fila de las unidades del dataframe
        unidades = unidades[0]  # Se redefine la variable para corregir ls dimensiones del array
        df = df[1:]  # elimina las filas con letras
        matriz = df.to_numpy()  # convierte el dataframe a numpy
        leyenda = np.array(list(df.columns))  # extrae el nombre de las columnas, sera la leyenda de las graficas
        matriz = matriz.transpose()

        self.tiempo.v = matriz[0]
        self.tiempo.l = leyenda[0]
        self.tiempo.u = unidades[0]
        self.aceleracion.v = matriz[1:-1]
        self.aceleracion.l = leyenda[1:-1]
        self.aceleracion.u = unidades[1:-1]
        self.fuerza.v = matriz[-1]
        self.fuerza.l = leyenda[-1]
        self.fuerza.u = unidades[-1]
        self.samplerate = self.comprobar_samplerate()
        self.comprobar_datos()
        self.axis_0, self.axis_1 = self.aceleracion.v.shape

    def escritura(self, ruta, nombre_archivo):
        # TODO: Comprobar que funciona correctamente
        leyenda = np.append(self.tiempo.l, self.aceleracion.l)
        leyenda = np.append(leyenda, self.fuerza.l)
        unidades = np.append(self.tiempo.u, self.aceleracion.u)
        unidades = np.append(unidades, self.fuerza.u)
        valores = np.append(self.tiempo.v, self.aceleracion.v)
        valores = np.append(valores, self.fuerza.v)

        df1 = pd.DataFrame(unidades, columns=leyenda)  # crea un dataframe solo con las unidades
        df2 = pd.DataFrame(valores.transpose(), columns=leyenda)  # crea un dataframe solo con los datos de los vectores

        df3 = df1.append(df2)  # Junta ambos dataframes, obteniendo un documento con el mismo formato de lectura.

        df3.to_excel(nombre_archivo + '/' + ruta,
                     index=False)  # guarda el documento en la ruta de "archivo", pueden especificarse en que hojas se guarde. Index=false no muestra el numero de las filas (tal como vienen los datos originalmente)

    def get_h5py(self,filepath):
        def undefined_to_array(value):
            if type(value) is np.ndarray:
                caracter = value[0]
                return caracter
            if type(value) is np.str_:
                caracter = str(value)
                return caracter
            return value

        try:
            os.remove(filepath)  # todo cambiar session
        except:
            pass

        self.fuerza.l = undefined_to_array(self.fuerza.l)

        with h5.File(filepath, 'w') as f:
            f.create_dataset('tiempo.tiempo.v', data=self.tiempo.v.astype(np.float64))
            f.create_dataset('tiempo.tiempo.u', data=self.tiempo.u)
            f.create_dataset('tiempo.tiempo.l', data=self.tiempo.l.astype('S'))
            f.create_dataset('tiempo.aceleracion.v', data=self.aceleracion.v.astype(np.float32))
            f.create_dataset('tiempo.aceleracion.u', data=self.aceleracion.u)
            f.create_dataset('tiempo.aceleracion.l', data=self.aceleracion.l.astype('S'))
            f.create_dataset('tiempo.fuerza.v', data=self.fuerza.v.astype(np.float64))
            f.create_dataset('tiempo.fuerza.u', data=self.fuerza.u)
            f.create_dataset('tiempo.fuerza.l', data=self.fuerza.l)
            f.create_dataset('tiempo.rms.v', data=self.rms.v)
            f.create_dataset('tiempo.rms_movil.v', data=self.rms_movil.v.astype(np.float64))
            f.create_dataset('tiempo.rms_movil_tiempo.v', data=self.tiempo_rms_movil.v.astype(np.float64))
            f.create_dataset('tiempo.mtvv', data=self.mtvv.v.astype(np.float64))
            f.create_dataset('tiempo.vdv.v', data=self.vdv.v.astype(np.float64))
            f.create_dataset('tiempo.vdv_tiempo.v', data=self.vdv.v.astype(np.float64))

    def from_h5py(self, filepath):
        with h5.File(filepath, 'r') as f:
            buffer = f.get('tiempo.tiempo.v')
            self.tiempo.v = np.array(buffer)
            self.tiempo.u = f.get('tiempo.tiempo.u').asstr()[()]
            self.tiempo.l = np.array(f.get('tiempo.tiempo.l').asstr()[()])

            buffer = f.get('tiempo.aceleracion.v')
            self.aceleracion.v = np.array(buffer)
            self.aceleracion.u = f.get('tiempo.aceleracion.u').asstr()[()]
            self.aceleracion.l = f.get('tiempo.aceleracion.l').asstr()[()]

            buffer = f.get('tiempo.fuerza.v')
            self.fuerza.v = np.array(buffer)
            self.fuerza.u = f.get('tiempo.fuerza.u').asstr()[()]
            self.fuerza.l = f.get('tiempo.fuerza.l').asstr()[()]

            buffer = f.get('tiempo.rms.v')
            self.rms.v = np.array(buffer)
            buffer = f.get('tiempo.rms_movil.v')
            self.rms_movil.v = np.array(buffer)
            buffer = f.get('tiempo.tiempo_rms_movil.v')
            self.tiempo_rms_movil.v = np.array(buffer)
            buffer = f.get('tiempo.mtvv.v')
            self.mtvv.v = np.array(buffer)
            buffer = f.get('tiempo.vdv.v')
            self.vdv.v = np.array(buffer)
            buffer = f.get('tiempo.tiempo_vdv.v')
            self.tiempo_vdv.v = np.array(buffer)

            self.rms.u = self.aceleracion.u
            self.rms_movil.u = self.aceleracion.u
            self.tiempo_rms_movil.u = self.tiempo.u
            self.mtvv.u = self.aceleracion.u
            self.vdv.u = self.aceleracion.u
            self.tiempo_vdv.u = self.tiempo.u

            self.rms.l = self.aceleracion.l
            self.rms_movil.l = self.aceleracion.l
            self.tiempo_rms_movil.l = self.tiempo.l
            self.mtvv.l = self.aceleracion.l
            self.vdv.l = self.aceleracion.l
            self.tiempo_vdv.l = self.tiempo.l
            self.comprobar_samplerate()
            self.axis_0, self.axis_1 = self.aceleracion.v.shape



    def calcular(self):
        self.get_rms_movil()
        self.get_rms()
        self.get_vdv()
        self.get_mtvv()

    def get_rms(self):
        self.rms = Atributo(valor=None, leyenda=self.aceleracion.l, unidad=self.aceleracion.u)
        self.rms.v = np.empty(self.axis_0)
        for i in range(self.axis_0):
            self.rms.v[i] = np.sqrt(np.mean(self.aceleracion.v[i] ** 2))

    def get_rms_movil(self, avance_rms = None, ventana_rms = None):
        self.tiempo_rms_movil = Atributo(valor=None, leyenda=self.aceleracion.l, unidad=self.aceleracion.u)
        self.rms_movil = Atributo(valor=None, leyenda=self.aceleracion.l, unidad=self.aceleracion.u)

        if self.samplerate is None:
            self.comprobar_samplerate()
        terminos_deseados = self.samplerate*preset.ventana_rms

        if terminos_deseados > self.axis_1:
            terminos_deseados = 5

        if avance_rms is None:
            avance_rms = round((1 - preset.solape_rms) * terminos_deseados)

        ventana = terminos_deseados - 1  # La ventana es el numero que hay que sumar al indice, mientas que los terminos deseados es el numero de terminos que tendra el vector

        tamano = (self.axis_1 + avance_rms - terminos_deseados) // avance_rms  # tamaño del vector de datos, es necesario definir antes el tamaño
        self.rms_movil.v = np.empty([self.axis_0, tamano])  # define el tamaño de la matriz de datos
        self.tiempo_rms_movil.v = np.empty([tamano])
        for i in range(tamano):
            self.tiempo_rms_movil.v[i] = (self.tiempo.v[avance_rms * i] + self.tiempo.v[
                ventana + avance_rms * i]) / 2  # Bucle que copia los datos del vector de tiempo original al nuevo, de menor tamaño

        for j in range(self.axis_0):
            for i in range(tamano):
                cache = self.aceleracion.v[j, (avance_rms * i):(ventana + 1 + avance_rms * i)]
                self.rms_movil.v[j, i] = np.sqrt(np.mean(
                    cache ** 2))  # Se efectua la formula de rms en todo el vector, iterando primero por los diferentes vectores y despues por los subvectores

    def get_mtvv(self):
        self.mtvv = Atributo(valor=None, leyenda=self.aceleracion.l, unidad=self.aceleracion.u)

        if self.rms_movil.v is None:
            self.get_rms_movil()
        self.mtvv.v = np.amax(self.rms_movil.v, axis=1)  # devuelve un vector con los maximos del rms, se obtiene un valor por cada vector

    def get_vdv(self):  # El incremento es el valor cada cuanto se tomaran datos para despies hacer la grafica, simplemente representa la precision de la misma.
        self.vdv = Atributo(valor=None, leyenda=self.aceleracion.l, unidad=self.aceleracion.u)
        self.tiempo_vdv = Atributo(valor=None, leyenda=self.aceleracion.l, unidad=self.aceleracion.u)

        tamano = self.axis_1 // preset.incremento_vdv
        self.tiempo_vdv.v = np.zeros(tamano)
        self.vdv.v = np.zeros([self.axis_0, tamano])  # crea una matriz de dimensiones correspondientes
        for i in range(tamano):
            self.tiempo_vdv.v[i] = self.tiempo.v[preset.incremento_vdv * i]  # Obtenemos el vector de tiempo

        for i in range(self.axis_0):
            sumandos = 0
            for j in range(tamano):  # doble bucle, por filas y columnas
                sumandos += np.sum((self.aceleracion.v[i, (preset.incremento_vdv * j):(preset.incremento_vdv * (j + 1))]) ** 4)
                self.vdv.v[i, j] = sumandos ** 0.25  # halla el vdv acumulado por cada acelerometro


    def get_frf(self, impact = True):
        def frecrate_entero(nperseg, noverlap, nfft, samplerate):
            if int(samplerate) != samplerate:
                return nperseg, noverlap, nfft
            if int(nfft) != nfft:
                return nperseg, noverlap, nfft

            def check(numero, decimal):
                caracter = str(numero)
                car_decimal = caracter.split('.')[1]
                if len(car_decimal) <= int(decimal):
                    return True
                else:
                    return False

            i = 0
            while True:
                frecrate = samplerate / nfft
                if check(frecrate, 6) is False:
                    i += 1
                    nfft += 1
                    if i == 10:
                        return nperseg, noverlap, nfft-10
                else:
                    return nperseg + i, noverlap + i, nfft

        def impacto_a_nperseg(fuerza, aceleracion, nperseg):
            impactos = np.array([])

            if abs(np.amax(fuerza))<abs(np.amin(fuerza)):
                fuerza = -fuerza

            max_fuerza = np.amax(fuerza)
            threshold = preset.threshold_frf * max_fuerza / 100
            for i in range(len(fuerza)-1):
                if fuerza[i]<threshold and fuerza[i+1]>= threshold:
                    impactos = np.append(impactos,i+1)

            impactos = impactos - 2
            impactos = impactos.astype(int)

            if len(impactos) == 0:
                raise ValueError('No hay impactos')

            if -1 in np.sign(impactos):
                raise ValueError('Impacto al comienzo de la senal')


            for i in range(len(impactos)-1):
                if impactos[i]+nperseg > impactos[i + 1]:
                    raise ValueError('El tiempo entre impactos es menor que nfft, diminuya nfft')
            if impactos[-1]+nperseg > self.axis_1:
                raise ValueError('nfft sobrepasa la longitud de la senal')

            fuerza_mod = np.array(())
            aceleracion_mod = np.empty([self.axis_0,0])
            for i in impactos:
                buffer = signal.detrend(fuerza[i:i + nperseg])
                buffer[0] = 0
                fuerza_mod = np.append(fuerza_mod, buffer)
                buffer = signal.detrend(aceleracion[:,i:i + nperseg])
                buffer[:,0] = 0
                aceleracion_mod = np.concatenate([aceleracion_mod, buffer], axis= 1)


            return fuerza_mod, aceleracion_mod

        if self.samplerate is None:
            self.comprobar_samplerate()
        nperseg = preset.nperseg_frf
        noverlap = preset.noverlap_frf
        nfft = preset.nfft_frf
        nperseg, noverlap, nfft = frecrate_entero(nperseg, noverlap, nfft, self.samplerate)

        if impact == True:
            fuerza,aceleracion = impacto_a_nperseg(self.fuerza.v, self.aceleracion.v, nperseg)
        else:
            fuerza, aceleracion = self.fuerza.v, self.aceleracion.v

        frec_1, Pxy = signal.csd(aceleracion, fuerza, fs=self.samplerate, window=preset.ventana_frf, nperseg=nperseg,
                                 noverlap=noverlap, nfft=nfft, detrend=False)
        # frec_1, Pxy = signal.csd(datos.fuerza.valor,datos.aceleracion.valor,fs= datos.samplerate,window = window,nperseg=nperseg,noverlap=noverlap,nfft=nfft)
        frec_2, Pxx = signal.welch(fuerza, fs=self.samplerate, window=preset.ventana_frf, nperseg=nperseg, noverlap=noverlap,
                                   nfft= nfft, detrend=False)
        if np.all(frec_1 != frec_2) is False:
            raise ValueError('No coinciden las frecuencias')

        # frfrecuencia = Pxy
        frfrecuencia = Pxy / Pxx


        unidad = self.aceleracion.u
        for i in range(len(unidad)):
            unidad[i] = str(unidad[i]+'/' + str(self.fuerza.u[0]))

        return Frecuencia(complejo=frfrecuencia, leyenda=self.aceleracion.l,
                          unidad=unidad, frecuencia=frec_1)

    def resample(self):
        self.comprobar_samplerate()
        # TODO: comprobar
        preset.samplerate_resample = int(preset.samplerate_resample)
        samples_iniciales = self.axis_1

        num_samples_decimal = samples_iniciales * preset.samplerate_resample / self.samplerate
        num_samples = math.trunc(num_samples_decimal)

        self.aceleracion.v, tiempo = signal.resample(self.aceleracion.v, num_samples, t=self.tiempo.v, axis=-1)
        self.fuerza.v, tiempo = signal.resample(self.fuerza.v, num_samples, t=self.tiempo.v, axis=-1)


        self.tiempo.v = np.linspace(0, (num_samples-1)/preset.samplerate_resample,num_samples)
        self.comprobar_samplerate()
        self.axis_1 = len(self.tiempo.v)

    def detrend(self):
        self.aceleracion.v = signal.detrend(self.aceleracion.v)

    def filtro(self):
        self.comprobar_samplerate()

        b, a = signal.butter(preset.orden_filtro, preset.frecuencia_filtro, 'low', fs=self.samplerate)

        self.aceleracion.v = signal.lfilter(b, a, self.aceleracion.v)
        self.fuerza.v = signal.lfilter(b, a, self.fuerza.v)

class Frecuencia:
    def __init__(self, complejo=None, magnitud=None, fase=None, fase_corr= None, real= None ,imag= None, leyenda=None, unidad=None, frecrate=None,
                 frecuencia=None, leyenda_frec='Frecuencia', unidad_frec='Hz'):
        self.complejo = Atributo(complejo, leyenda, unidad)
        self.magnitud = Atributo(magnitud, leyenda, unidad)
        self.fase = Atributo(fase, leyenda, 'rad')
        self.fase_corr = Atributo(fase_corr, leyenda, 'rad')
        self.real = Atributo(real, leyenda, unidad)
        self.imag = Atributo(imag, leyenda, unidad)
        self.frecuencia = Atributo(frecuencia, leyenda_frec, unidad_frec)

        if self.complejo.v is not None:
            self.axis_0, self.axis_1 = self.complejo.v.shape  # axis_0 representa el numero de acelerometros y axis_1 el numero de medidas


        if (real is None or imag is None) and self.complejo.v is not None:
            self.get_real_imag()

        if (magnitud is None or fase is None) and self.complejo.v is not None:
            self.get_polares()

        if frecrate is None and self.frecuencia.v is not None:
            self.get_frecrate()
        else:
            self.frecrate = frecrate

        if self.fase.v is not None and fase_corr is None:
            self.corregir_fase()


    def get_h5py(self, filepath):
        try:
            os.remove(filepath)  # todo cambiar session
        except:
            pass


        with h5.File(filepath, 'w') as f:
            f.create_dataset('magnitud', data=self.magnitud.v.astype(np.float64))
            f.create_dataset('fase', data=self.fase.v.astype(np.float64))
            f.create_dataset('unidad', data=self.real.u)
            f.create_dataset('leyenda', data=self.real.l.astype('S'))

            f.create_dataset('frecuencia', data=self.frecuencia.v.astype(np.float64))
            f.create_dataset('unidad_frec', data=self.frecuencia.u)
            f.create_dataset('leyenda_frec', data=self.frecuencia.l)

            f.create_dataset('fase_corr', data=self.fase_corr.v.astype(np.float64))
            f.create_dataset('real', data=self.real.v.astype(np.float64))
            f.create_dataset('imag', data=self.imag.v.astype(np.float64))
            f.create_dataset('frecrate', data=self.frecrate)


    def from_h5py(self, filepath):

        with h5.File(filepath, 'r') as f:

            buffer = f.get('magnitud')
            self.magnitud.v = np.array(buffer)
            buffer = f.get('fase')
            self.fase.v = np.array(buffer)
            self.real.u = f.get('unidad').asstr()[()]
            self.real.l = f.get('leyenda').asstr()[()]

            buffer = f.get('frecuencia')
            self.frecuencia.v = np.array(buffer)
            self.frecuencia.u = f.get('unidad_frec').asstr()[()]
            self.frecuencia.l = f.get('leyenda_frec').asstr()[()]

            buffer = f.get('fase_corr')
            self.fase_corr.v = np.array(buffer)
            buffer = f.get('real')
            self.real.v = np.array(buffer)
            buffer = f.get('imag')
            self.imag.v = np.array(buffer)
            buffer = f.get('frecrate')
            self.frecrate = np.array(buffer)

            self.get_cartesianas()
            self.axis_0, self.axis_1 = self.complejo.v.shape  # axis_0 representa el numero de acelerometros y axis_1 el numero de medidas

            self.complejo.u = self.real.u
            self.magnitud.u = self.real.u
            self.fase.u = 'rad'
            self.fase_corr.u = 'rad'
            self.imag.u = self.real.u

            self.complejo.l = self.real.l
            self.magnitud.l = self.real.l
            self.fase.l = self.real.l
            self.fase_corr.l = self.real.l
            self.imag.l = self.real.l

    def get_polares(self):
        # TODO: añadir magnitud en dB
        self.magnitud.v = abs(self.complejo.v)

        axis_0, axis_1 = self.complejo.v.shape
        self.magnitud.v = np.empty([axis_0,
                                    axis_1])  # se crean dos vectores, modulo y fase, para transformar los numeros a coordenadas polares
        self.fase.v = np.empty([axis_0, axis_1])

        for i in range(
                axis_0):  # la funcion cmath.polar no trabaja con vectores, es necesario iterar por los vectores para obtener modulo y fase (en radianes)
            for j in range(axis_1):
                self.magnitud.v[i, j], self.fase.v[i, j] = cmath.polar(self.complejo.v[i, j])

    def get_cartesianas(self):
        self.complejo.v = self.magnitud.v * np.exp(1j * self.fase.v)

    def get_real_imag(self):
        axis_0, axis_1 = self.complejo.v.shape
        self.real.v = np.empty([axis_0,
                                axis_1])  # se crean dos vectores, modulo y fase, para transformar los numeros a coordenadas polares
        self.imag.v = np.empty([axis_0, axis_1])

        for i in range(
                axis_0):  # la funcion cmath.polar no trabaja con vectores, es necesario iterar por los vectores para obtener modulo y fase (en radianes)
            for j in range(axis_1):
                self.real.v[i, j] = self.complejo.v[i, j].real
                self.imag.v[i, j] = self.complejo.v[i, j].imag

    def corregir_fase(self):
        self.fase_corr = Atributo(self.fase.v, self.fase.l, self.fase.u)
        for i in range(self.axis_0):
            for j in range(self.axis_1 - 1):
                k = 1
                while True:
                    if self.frecuencia.v[j] <= 1.0:
                        break
                    valor = self.fase_corr.v[i, j + 1] - self.fase_corr.v[i, j]
                    if abs(valor) < np.pi + 0.001:
                        break
                    if valor < 0:
                        self.fase_corr.v[i, j + 1] += (np.pi * 2)

                    else:
                        self.fase_corr.v[i, j + 1] -= (np.pi * 2)
                    if k > 30:
                        break
                    k += 1

    def get_frecrate(self):
        # Se comprueba primero el samplerate de manera precisa, contando el numero
        # de iteraciones sobre el vector tiempo hasta alcanzar el valor 1.
        # En caso de ser un vector de longitud suficiente retorna el samplerate exacto, samples
        samples = 0
        t = 0
        while t < 1:
            samples += 1
            if samples == len(self.frecuencia.v):
                break
            t = self.frecuencia.v[samples]
            if t == 1:
                self.frecrate = samples
        # Si el metodo anterior falla se recurre a hallar el samplerate dividiendo 1 entre el incremento
        # de tiempo dt = tiempo[1] - tiempo [0]. Si falla se retorna el valor más cercano al samplerate
        # con una indicación de error para que el usuario revise los datos

        samplerate_decimal = 1 / (self.frecuencia.v[1] - self.frecuencia.v[0])
        samplerate = samplerate_decimal

        self.frecrate = samplerate


    def copia(self):
        return self.complejo.v, self.magnitud.v, self.fase.v, self.complejo.l, \
               self.complejo.u, self.frecrate, self.frecuencia.v, \
               self.frecuencia.l, self.frecuencia.u

    def integrar(self):
        self.fase.v = self.fase.v + np.pi / 2
        self.fase_corr.v = self.fase_corr.v + np.pi / 2
        frecuencia_mod = self.frecuencia.v
        frecuencia_mod[0] = 1
        self.magnitud.v = self.magnitud.v / frecuencia_mod
        self.frecuencia.v[0] = 0
        self.get_cartesianas()
        self.get_real_imag()

class Modal:
    def __init__(self, data, rango_frec= None, rango_unico= None):

        self.frf = data

        if rango_frec is not None:
            self.rango_frec = rango_frec
        elif rango_unico is not None:
            self.adaptar_rango(rango_unico)
        else:
            raise ValueError('Se debe introducir un rango de frecuencias')
        shape = self.rango_frec.shape
        self.axis_0 = shape[0]
        self.axis_1 = shape[1]

        self.frec_to_place()

        self.frec_resonancia = np.zeros((self.axis_0,self.axis_1))
        self.amortiguamiento = np.zeros((self.axis_0,self.axis_1))

    def adaptar_rango(self, rango):
        rango = rango.tolist()
        lista = list()
        for i in range(self.frf.axis_0):
            lista.append(rango)
        array = [*lista]
        self.rango_frec = np.array(array)

        shape = self.rango_frec.shape
        self.axis_0 = shape[0]
        self.axis_1 = shape[1]
        self.frec_to_place()

    def frec_to_place(self):
        self.pos_rango_frec = np.zeros(self.rango_frec.shape)

        for i in range(self.axis_0):
            for j in range(self.axis_1):
                for k in range(2):
                    buffer = self.rango_frec[i,j,k] * self.frf.frecrate
                    self.pos_rango_frec[i,j,k] = int(round(buffer))

        self.pos_rango_frec = self.pos_rango_frec.astype(np.int)

    def peak_picking_cross(self, rango,Pot_mitad):

        pos = np.empty(0)
        buffer = rango - Pot_mitad


        for i in range(len(rango)-1):
            signo = np.sign(buffer[i])
            if signo != np.sign(buffer[i+1]):
                j = i + abs(buffer[i])/abs(buffer[i]-buffer[i+1])

                pos = np.append(pos,j)
            if buffer[i] == 0:
                pos = np.append(pos,i)
        if len(pos) == 0:
            return 0
        return np.mean(pos)

    def peak_picking(self):
        # nombrar variables
        self.frec_resonancia = np.zeros([self.axis_0,self.axis_1])
        self.amortiguamiento = np.zeros([self.axis_0,self.axis_1])
        # bucle principal
        for i in range(self.axis_0):
            for j in range(self.axis_1):
                self.peak_picking_single(i,j)

    def peak_picking_single(self,i,j):

        frec_baja = self.pos_rango_frec[i, j, 0]
        frec_alta = self.pos_rango_frec[i, j, 1]
        # Detectamos en que valor del vector frecuencias se da el maximo:


        pos_wn = np.argmax(self.frf.magnitud.v[i, frec_baja:frec_alta])
        pos_wn = pos_wn + frec_baja
        wn = self.frf.frecuencia.v[pos_wn]
        self.frec_resonancia[i, j] = wn

        if frec_baja == pos_wn or frec_alta == pos_wn:
            self.frec_resonancia[i,j] = 0
            self.amortiguamiento[i,j] = 0
            return None
        # Se calcula la potencia mitad del pico, y se buscan los valores que coinciden con ella.
        pot_mitad = self.frf.magnitud.v[i, pos_wn] / np.sqrt(2)

        media = self.peak_picking_cross(self.frf.magnitud.v[i, frec_baja:(pos_wn + 1)], pot_mitad)
        if media == 0:
            self.frec_resonancia[i,j] = 0
            self.amortiguamiento[i,j] = 0
            return None
        w1 = (self.pos_rango_frec[i, j, 0]+media)/self.frf.frecrate


        media = self.peak_picking_cross(self.frf.magnitud.v[i, pos_wn:frec_alta], pot_mitad)
        if media == 0:
            self.frec_resonancia[i,j] = 0
            self.amortiguamiento[i,j] = 0
            return None
        w2 = (pos_wn+media)/self.frf.frecrate

        self.amortiguamiento[i, j] = (w2 - w1)/(2*wn)

    def circle_fit(self, debug_sp = False, debug_nyquist= False, debug_psi = False):
        def single_fit(x,y,frecrate,n0,i):
            def centro(x, y):
                def estimar(x, y):
                    x_m = np.mean(x)
                    y_m = np.mean(y)
                    return np.array([x_m, y_m])

                def calc_R(xc, yc):
                    """ calculate the distance of each 2D points from the center (xc, yc) """
                    return np.sqrt((x - xc) ** 2 + (y - yc) ** 2)

                def f_2(c):
                    """ calculate the algebraic distance between the data points and the mean circle centered at c=(xc, yc) """
                    Ri = calc_R(*c)
                    return Ri - Ri.mean()

                center_estimate = estimar(x, y)
                center_2, ier = optimize.leastsq(f_2, center_estimate)

                xc_2, yc_2 = center_2
                Ri_2 = calc_R(*center_2)
                R_2 = Ri_2.mean()
                return xc_2, yc_2, R_2

            def espacio(x, y, x_c, y_c, i):

                def corte(rango, Pot_mitad):
                    pos = np.empty(0)
                    buffer = rango - Pot_mitad
                    for i in range(len(rango) - 1):
                        signo = np.sign(buffer[i])
                        if signo != np.sign(buffer[i + 1]):
                            j = i + abs(rango[i]) / abs(rango[i] - rango[i + 1])
                            pos = np.append(pos, j)
                        if buffer[i] == 0:
                            pos = np.append(pos, i)
                    if len(pos) == 0:
                        return 0
                    return np.mean(pos)

                sp = np.arctan2((y - y_c) , (x - x_c))

                if debug_nyquist is True:
                    fig = go.Figure()
                    fig.add_scatter(x=x - x_c, y=y - y_c)
                    fig.add_scatter(x=[0], y=[0])
                    fig.update_layout(yaxis=dict(scaleanchor="x", scaleratio=1))
                    fig.write_html('ny-{}.html'.format(i))


                buffer = 0
                for j in range(len(x)-1):

                    if ((np.sign(x[j]- x_c) == -1 and np.sign(y[j] - y_c) == 1 and
                         np.sign(x[j+1]- x_c) == -1 and np.sign(y[j+1] - y_c) == -1)):
                        buffer += 2* np.pi
                        sp[j+1] += buffer

                    if ((np.sign(x[j]- x_c) == -1 and np.sign(y[j] - y_c) == -1 and
                         np.sign(x[j+1]- x_c) == -1 and np.sign(y[j+1] - y_c) == 1)):
                        buffer -= 2* np.pi
                        sp[j+1] += buffer
                    else:
                        sp[j+1] += buffer


                g_sp = np.gradient(sp)
                gg_sp = np.gradient(g_sp)

                max = np.argmax(abs(g_sp))

                punto = max - 4 + corte(gg_sp[max - 4:max + 4], 0)

                f = interpolate.interp1d(range(len(sp)), sp)

                try:
                    sp = sp - f(punto)
                except ValueError:
                    return False, False, False
                if debug_sp is True:
                    fig = go.Figure()
                    fig.add_scatter(x=np.array(range(len(sp))), y=sp)
                    fig.add_scatter(x=np.array(range(len(g_sp))), y=g_sp)
                    fig.add_scatter(x=np.array(range(len(gg_sp))), y=gg_sp)

                    fig.write_html('sp-{}.html'.format(i))
                return g_sp, sp, punto
                # return punto, sp

            def amor(sp, pos_wn, n0, frecrate):
                wa = (n0 + np.array(range(len(sp)))) / frecrate
                wr = (n0 + pos_wn) / frecrate
                if len(wa) == 0:
                    return 0,0

                Psi = ((wr ** 2 - wa ** 2) / (2 * wr * wa)) * (1 / (np.tan(sp / 2)))

                psi = np.mean(Psi[int(pos_wn)-2:int(pos_wn)+2])

                if debug_psi is True:
                    fig = go.Figure()
                    fig.add_scatter( x = wa, y=Psi)
                    fig.write_html('psi-{}.html'.format(i))

                if np.isnan(psi):
                    wr, psi = 0, 0
                if np.sign(psi) == -1:
                    wr, psi = 0, 0

                return wr, psi

            xc_2, yc_2, R_2 = centro(x, y)

            if (xc_2>np.amax(x) or xc_2<np.amin(x) or yc_2>np.amax(y) or yc_2<np.amin(y)):
                return 0,0

            g_sp, sp, pos_wn = espacio(x, y, xc_2, yc_2, i)

            if sp is False:
                return 0,0

            wr, psi = amor(sp, pos_wn, n0, frecrate)

            return wr, psi


        # nombrar variables
        self.frec_resonancia = np.zeros([self.axis_0,self.axis_1])
        self.amortiguamiento = np.zeros([self.axis_0,self.axis_1])
        # bucle principal
        for i in range(self.axis_0):
            for j in range(self.axis_1):
                frec_baja = self.pos_rango_frec[i, j, 0]
                frec_alta = self.pos_rango_frec[i, j, 1]
                x = self.frf.imag.v[i,frec_baja:frec_alta]
                y = self.frf.real.v[i,frec_baja:frec_alta]
                n0 = self.pos_rango_frec[i,j,0]
                wr,psi = single_fit(x,y,self.frf.frecrate,n0,i)
                self.frec_resonancia[i, j] = wr
                self.amortiguamiento[i, j] = psi


    def promedio(self):
        amortiguamiento = np.transpose(self.amortiguamiento)
        self.amortiguamiento_medio = np.zeros(len(amortiguamiento))
        for i in range(len(amortiguamiento)):
            k = 0
            for j in range(len(amortiguamiento[0])):
                if amortiguamiento[i,j] != 0:
                    k += 1
            total = sum(amortiguamiento[i])
            self.amortiguamiento_medio[i] = total/k

        frec_resonancia = np.transpose(self.frec_resonancia)
        self.frec_resonancia_media = np.zeros(len(frec_resonancia))
        for i in range(len(frec_resonancia)):
            k = 0
            for j in range(len(frec_resonancia[i])):
                if frec_resonancia[i,j] != 0:
                    k += 1
            total = sum(frec_resonancia[i])
            self.frec_resonancia_media[i] = total/k



