from plotly.subplots import make_subplots
import plotly.graph_objects as go
import numpy as np

import os


template = 'presentation'
archivo = 'grafica.html'


def grafica(tiempo, *args):
    # https://plotly.com/python/legend/
    try:
        if len(args) == 1:
            arg = args[0]
            try:
                y_title = arg.u[0]
            except:
                y_title = arg.u
        else:
            y_title = ''
        x_title =  tiempo.u
    except TypeError:
        y_title = ' '
        x_title = ' '
    fig = go.Figure()
    for arg in args:
        if len(arg.v.shape) == 1:
            fig.add_trace(go.Scatter(x=tiempo.v, y=arg.v, name=arg.l, mode='lines', line=dict(width=1.5)))
        else:
            for i in range(len(arg.v)):
                fig.add_trace(go.Scatter(x = tiempo.v, y = arg.v[i], name = arg.l[i], mode='lines', line=dict(width=1.5)))
    fig.update_layout(title= None,
                      xaxis_title= x_title,
                      yaxis_title= y_title,
                      template= template)
    return fig

def grafica_html(tiempo, *args, arch = None):
    if arch is None:
        arch = archivo
    fig = grafica(tiempo, *args)
    return fig.write_html(arch)

def bode(data):
    print('data',data.magnitud.u)
    fig = make_subplots(
    rows=2, cols=1,
    subplot_titles=("Magnitud (dB con ref 1 {})".format(data.magnitud.u[0]),"Fase ({})".format(data.fase.u)))

    for i in range(data.axis_0):
        fig.add_trace(go.Scatter(x=data.frecuencia.v, y=20*np.log(data.magnitud.v[i]), name = data.magnitud.l[i], mode='lines', line=dict(width=1.5)),#todo change log data
                                 row=1, col=1)
        # fig.add_trace(go.Scatter(x=data.frecuencia.v, y=data.magnitud.v[i], name = data.magnitud.l[i], mode='lines'),#todo change log data
        # row=1, col=1)

        fig.add_trace(go.Scatter(x=data.frecuencia.v, y=data.fase_corr.v[i], name=data.fase.l[i], mode='lines', line=dict(width=1.5)),
                 row=2, col=1)

    fig.update_layout(template=template,
                      height = 800,
                      xaxis_title = 'Hz')
    return fig

def bode_html(data, arch = None):
    if arch is None:
        arch = archivo
    fig = bode(data)
    return fig.write_html(arch)

def nyquist(data):
    fig = go.Figure()

    for i in range(data.axis_0):
        fig.add_trace(go.Scatter(x= data.real.v[i], y=data.imag.v[i], name=data.real.l[i], mode='lines', line=dict(width=1.5)))

    fig.update_layout(title= None,
                      template= template,
                      xaxis_title= 'Real ({})'.format(data.real.u[0]),
                      yaxis_title= 'Imag ({})'.format(data.real.u[0]),
                      yaxis=dict(scaleanchor="x", scaleratio=1),
                      height= 900)

    return fig

def nyquist_html(data, arch = None):
    if arch is None:
        arch = archivo
    fig = nyquist(data)
    return fig.write_html(arch)
